.bk.opt {
    position: relative;
    display: block;
    left: 75px;
    top: 0px;
    width: 80px;
    height: 80px;
}

.bk.background {
    border-radius: 10px;
    background-color: rgba(0, 0, 0, 0.03);
}

.bk.upload_data {
    border-radius: 5px;
    border: 1px #fff solid;
    padding: 20px;
}

h1 {
    color: #045082;
    font-size: 45px;
    line-height: 0.6;
    text-align: center;
}

h2 {
    color: #045082;
    text-align: center;
}

.bk.main-part {
    background-color: #eaeaea;
    font-size: 17px;
    line-height: 23px;
    letter-spacing: 0px;
    font-weight: 500;
    color: #045082;
    text-align: center;
    position: relative !important;
    margin-left: auto;
    margin-right: auto;
    width: 40%;
}

.bk-root .bk-btn-primary {
    background-color: #045082;
    font-size: 12px;
    font-weight: 700;
    text-transform: uppercase;
    letter-spacing: 1.5px;
    font-weight: bold;
}

.bk-root .bk-btn-default {
    font-size: 11px;
    font-weight: 700;
    text-transform: uppercase;
    letter-spacing: 1.5px;
    background-color: #fff;
    /* border-color: #ccc; */
    height: 100%;
    display: inline-block;
    text-align: center;
    vertical-align: middle;
    white-space: nowrap;
    cursor: pointer;
    padding: 6px 12px;
    border: 1px solid #ccc;
    border-radius: 4px;
    outline: 0;
    user-select: none;
}

progress.light[value]::-webkit-progress-value {
  background-color: #7f7f7f;
  /* border-radius: 50px; */
}
progress.light:not([value])::before {
  background-color: #7f7f7f;
  /* border-radius: 50px; */
}
progress.light:not([value])::-webkit-progress-bar {
  background-color: #7f7f7f;
  /* border-radius: 50px; */
}
progress.light::-moz-progress-bar {
  background-color: #7f7f7f;
  /* border-radius: 50px; */
}

.bk.alert-danger {
    background-color: rgba(0, 0, 0, 0);
    color: #c72d3b;
    border: 0px #eaeaea solid;
    padding: 0;
}

.bk.alert-success {
    background-color: rgba(0, 0, 0, 0.03);
    border: 0px #eaeaea solid;
    padding: 0;
}

.settings {
    background-color: #eaeaea;
    border: 2px solid #045082;
}

.bk.card-header {
    background-color: #eaeaea;
    color: #333 !important;;
    text-align: center;
}

.bk-root .bk-tabs-header {
    background-color: #eaeaea;
    color: #333;
    border: 2px solid #eaeaea;
    font-size: 1.3em;
    font-weight: bold;
    padding: 3px;
    text-align: center;
    border-radius: 3px;
    /* text-transform: uppercase; */
}

.bk-root .bk-tabs-header .bk-tab{
    color: #333 !important;
}

.bk-root .bk-tabs-header .bk-tab.bk-active {
    color: #045082 !important;
    background-color: white;
}

.tabulator .tabulator-header .tabulator-col {
    background-color: #444;
}

.tabulator .tabulator-header {
    background-color: #444;
    color: white;
    border-bottom: 3px solid #045082;
}

a:link, a:visited {
  color: blue;
}
